:host(.bk-panel-models-modal-Modal) {
  width: 0;
  height: 0;
}
.dialog-container,
.dialog-overlay {
  position: fixed;
  top: 0;
  right: 0;
  bottom: 0;
  left: 0;
}
.dialog-container {
  z-index: 100002;
  display: flex;
}
.dialog-overlay {
  z-index: 100001;
  background-color: rgba(43, 46, 56, 0.9);
}
.dialog-content {
  margin: auto;
  z-index: 100002;
  position: relative;
  background-color: white;
  border-radius: 2px;
  padding: 10px;
  padding-bottom: 20px;
}
@keyframes fade-in {
  from {
    opacity: 0;
  }
}
@keyframes slide-up {
  from {
    transform: translateY(10%);
  }
}
.dialog-overlay {
  animation: fade-in 200ms both;
}
.dialog-content {
  animation: fade-in 400ms 200ms both, slide-up 400ms 200ms both;
}
@media (prefers-reduced-motion: reduce) {
  .dialog-overlay,
  .dialog-content {
    animation: none;
  }
}
.pnx-dialog-close {
  position: absolute;
  top: 0;
  right: 0;
  border: 0;
  padding: 0;
  background-color: transparent;
  font-size: 1.5em;
  width: 1.5em;
  height: 1.5em;
  text-align: center;
  cursor: pointer;
  transition: 0.15s;
  border-radius: 50%;
  z-index: 100003;
}
.pnx-dialog-close:hover {
  background-color: rgba(50, 50, 0, 0.15);
}
.lm-Widget.p-Widget.lm-TabBar.p-TabBar.lm-DockPanel-tabBar.jp-Activity {
  z-index: -1;
}
